# This file borrowed heavily from:
# https://github.com/mozilla-services/macauthlib

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import unittest

from webob import Request

from pyramid_jwtauth.utils import (strings_differ,
                                   parse_authz_header)


class TestUtils(unittest.TestCase):

    def test_strings_differ(self):
        # We can't really test the timing-invariance, but
        # we can test that we actually compute equality!
        self.assertTrue(strings_differ("", "a"))
        self.assertTrue(strings_differ("b", "a"))
        self.assertTrue(strings_differ("cc", "a"))
        self.assertTrue(strings_differ("cc", "aa"))
        self.assertFalse(strings_differ("", ""))
        self.assertFalse(strings_differ("D", "D"))
        self.assertFalse(strings_differ("EEE", "EEE"))

    def test_parse_authz_header(self):
        def req(authz):
            """Make a fake request with the given authz header."""
            class request:
                environ = {"HTTP_AUTHORIZATION": authz}
            return request

        # Test parsing of a single unquoted parameter.
        params = parse_authz_header(req('Digest realm=hello'))
        self.assertEqual(params['scheme'], 'Digest')
        self.assertEqual(params['realm'], 'hello')

        # Test parsing of multiple parameters with mixed quotes.
        params = parse_authz_header(req('Digest test=one, again="two"'))
        self.assertEqual(params['scheme'], 'Digest')
        self.assertEqual(params['test'], 'one')
        self.assertEqual(params['again'], 'two')

        # Test parsing of an escaped quote and empty string.
        params = parse_authz_header(req('Digest test="\\"",again=""'))
        self.assertEqual(params['scheme'], 'Digest')
        self.assertEqual(params['test'], '"')
        self.assertEqual(params['again'], '')

        # Test parsing of embedded commas, escaped and non-escaped.
        params = parse_authz_header(req('Digest one="1\\,2", two="3,4"'))
        self.assertEqual(params['scheme'], 'Digest')
        self.assertEqual(params['one'], '1,2')
        self.assertEqual(params['two'], '3,4')

        # Test parsing on various malformed inputs
        self.assertRaises(ValueError, parse_authz_header, req(None))
        self.assertRaises(ValueError, parse_authz_header, req(""))
        self.assertRaises(ValueError, parse_authz_header, req(" "))
        self.assertRaises(ValueError, parse_authz_header,
                          req('Broken raw-token'))
        self.assertRaises(ValueError, parse_authz_header,
                          req('Broken realm="unclosed-quote'))
        self.assertRaises(ValueError, parse_authz_header,
                          req('Broken realm=unopened-quote"'))
        self.assertRaises(ValueError, parse_authz_header,
                          req('Broken realm="unescaped"quote"'))
        self.assertRaises(ValueError, parse_authz_header,
                          req('Broken realm="escaped-end-quote\\"'))
        self.assertRaises(ValueError, parse_authz_header,
                          req('Broken realm="duplicated",,what=comma'))

        # Test all those again, but returning a default value
        self.assertEqual(None, parse_authz_header(req(None), None))
        self.assertEqual(None, parse_authz_header(req(""), None))
        self.assertEqual(None, parse_authz_header(req(" "), None))
        self.assertEqual(None,
                          parse_authz_header(req('Broken raw-token'), None))
        self.assertEqual(None, parse_authz_header(
                          req('Broken realm="unclosed-quote'), None))
        self.assertEqual(None, parse_authz_header(
                          req('Broken realm=unopened-quote"'), None))
        self.assertEqual(None, parse_authz_header(
                          req('Broken realm="unescaped"quote"'), None))
        self.assertEqual(None, parse_authz_header(
                          req('Broken realm="escaped-end-quote\\"'), None))
        self.assertEqual(None, parse_authz_header(
                          req('Broken realm="duplicated",,what=comma'), None))




# utils for pyramid_jwtauth

# This code heavily borrowed from:
# https://github.com/mozilla-services/macauthlib

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
"""

Low-level utility functions for pyramid_jwtauth

"""

import sys
import re
import functools
import base64
import webob

requests = None
try:
    import requests
except ImportError:   # pragma nocover
    pass

if sys.version_info > (3,):  # pragma: nocover

    def iteritems(d):
        """Efficiently iterate over dict items."""
        return d.items()

    def b64encode(data):
        """Base64-encode bytes data into a native string."""
        return base64.b64encode(data).decode("ascii")

else:  # pragma: nocover

    def iteritems(d):  # NOQA
        """Efficiently iterate over dict items."""
        return d.iteritems()

    def b64encode(data):  # NOQA
        """Base64-encode bytes data into a native string."""
        return base64.b64encode(data)


# Regular expression matching a single param in the HTTP_AUTHORIZATION header.
# This is basically <name>=<value> where <value> can be an unquoted token,
# an empty quoted string, or a quoted string where the ending quote is *not*
# preceded by a backslash.
_AUTH_PARAM_RE = r'([a-zA-Z0-9_\-]+)=(([a-zA-Z0-9_\-]+)|("")|(".*[^\\]"))'
_AUTH_PARAM_RE = re.compile(r"^\s*" + _AUTH_PARAM_RE + r"\s*$")

# Regular expression matching an unescaped quote character.
_UNESC_QUOTE_RE = r'(^")|([^\\]")'
_UNESC_QUOTE_RE = re.compile(_UNESC_QUOTE_RE)

# Regular expression matching a backslash-escaped characer.
_ESCAPED_CHAR = re.compile(r"\\.")


def parse_authz_header(request, *default):
    """Parse the authorization header into an identity dict.

    This function can be used to extract the Authorization header from a
    request and parse it into a dict of its constituent parameters.  The
    auth scheme name will be included under the key "scheme", and any other
    auth params will appear as keys in the dictionary.

    For example, given the following auth header value:

        'Digest realm="Sync" userame=user1 response="123456"'

    This function will return the following dict:

        {"scheme": "Digest", realm: "Sync",
         "username": "user1", "response": "123456"}

    """
    # This outer try-except catches ValueError and
    # turns it into return-default if necessary.
    try:
        # Grab the auth header from the request, if any.
        authz = request.environ.get("HTTP_AUTHORIZATION")
        if authz is None:
            raise ValueError("Missing auth parameters")
        scheme, kvpairs_str = authz.split(None, 1)
        # Split the parameters string into individual key=value pairs.
        # In the simple case we can just split by commas to get each pair.
        # Unfortunately this will break if one of the values contains a comma.
        # So if we find a component that isn't a well-formed key=value pair,
        # then we stitch bits back onto the end of it until it is.
        kvpairs = []
        if kvpairs_str:
            for kvpair in kvpairs_str.split(","):
                if not kvpairs or _AUTH_PARAM_RE.match(kvpairs[-1]):
                    kvpairs.append(kvpair)
                else:
                    kvpairs[-1] = kvpairs[-1] + "," + kvpair
            if not _AUTH_PARAM_RE.match(kvpairs[-1]):
                raise ValueError('Malformed auth parameters')
        # Now we can just split by the equal-sign to get each key and value.
        params = {"scheme": scheme}
        for kvpair in kvpairs:
            (key, value) = kvpair.strip().split("=", 1)
            # For quoted strings, remove quotes and backslash-escapes.
            if value.startswith('"'):
                value = value[1:-1]
                if _UNESC_QUOTE_RE.search(value):
                    raise ValueError("Unescaped quote in quoted-string")
                value = _ESCAPED_CHAR.sub(lambda m: m.group(0)[1], value)
            params[key] = value
        return params
    except ValueError:
        if default:
            return default[0]
        raise


def strings_differ(string1, string2):
    """Check whether two strings differ while avoiding timing attacks.

    This function returns True if the given strings differ and False
    if they are equal.  It's careful not to leak information about *where*
    they differ as a result of its running time, which can be very important
    to avoid certain timing-related crypto attacks:

        http://seb.dbzteam.org/crypto/python-oauth-timing-hmac.pdf

    """
    if len(string1) != len(string2):
        return True
    invalid_bits = 0
    for a, b in zip(string1, string2):
        invalid_bits += ord(a) ^ ord(b)
    return invalid_bits != 0


def normalize_request_object(func):
    """Decorator to normalize request into a WebOb request object.

    This decorator can be applied to any function taking a request object
    as its first argument, and will transparently convert other types of
    request object into a webob.Request instance.  Currently supported
    types for the request object are:

        * webob.Request objects
        * requests.Request objects
        * WSGI environ dicts
        * bytestrings containing request data
        * file-like objects containing request data

    If the input request object is mutable, then any changes that the wrapped
    function makes to the request headers will be written back to it at exit.
    """
    @functools.wraps(func)
    def wrapped_func(request, *args, **kwds):
        orig_request = request
        # Convert the incoming request object into a webob.Request.
        if isinstance(orig_request, webob.Request):
            pass
        # A requests.PreparedRequest object?
        elif requests and isinstance(orig_request, requests.PreparedRequest):
            # Copy over only the details needed for the signature.
            # WebOb doesn't code well with bytes header names,
            # so we have to be a little careful.
            request = webob.Request.blank(orig_request.url)
            request.method = orig_request.method
            for k, v in iteritems(orig_request.headers):
                if not isinstance(k, str):
                    k = k.decode('ascii')
                request.headers[k] = v
        # A WSGI environ dict?
        elif isinstance(orig_request, dict):
            request = webob.Request(orig_request)
        # A bytestring?
        elif isinstance(orig_request, bytes):
            request = webob.Request.from_bytes(orig_request)
        # A file-like object?
        elif all(hasattr(orig_request, attr) for attr in ("read", "readline")):
            request = webob.Request.from_file(orig_request)

        # The wrapped function might modify headers.
        # Write them back if the original request object is mutable.
        try:
            return func(request, *args, **kwds)
        finally:
            if requests and isinstance(orig_request, requests.PreparedRequest):
                orig_request.headers.update(request.headers)

    return wrapped_func



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import print_function

import unittest
import json
import time
import datetime

from webtest import TestApp

from zope.interface.verify import verifyClass

from pyramid.request import Request
from pyramid.response import Response
from pyramid.config import Configurator
from pyramid.interfaces import IAuthenticationPolicy
from pyramid.httpexceptions import HTTPForbidden, HTTPUnauthorized
from pyramid.security import (unauthenticated_userid,
                              authenticated_userid,
                              effective_principals,
                              Everyone,
                              Authenticated)

from pyramid_jwtauth import (JWTAuthenticationPolicy,
                             authenticate_request as jwt_authenticate_request)
import pyramid_jwtauth.utils

MASTER_SECRET = "V8 JUICE IS 1/8TH GASOLINE"


def make_request(config, path="/", environ={}):
    """Helper function for making pyramid Request objects."""
    my_environ = {}
    my_environ["wsgi.version"] = (1, 0)
    my_environ["wsgi.multithread"] = True
    my_environ["wsgi.multiprocess"] = True
    my_environ["wsgi.run_once"] = False
    my_environ["wsgi.url_scheme"] = "http"
    my_environ["REQUEST_METHOD"] = "GET"
    my_environ["SCRIPT_NAME"] = ""
    my_environ["PATH_INFO"] = path
    my_environ["SERVER_NAME"] = "localhost"
    my_environ["SERVER_PORT"] = "5000"
    my_environ["QUERY_STRING"] = "5000"
    my_environ.update(environ)
    request = Request(my_environ)
    request.registry = config.registry
    return request


# Something non-callable, to test loading non-callables by name.
stub_non_callable = None


def stub_find_groups(userid, request):
    """Groupfinder with the following rules:

        * any user with "bad" in their name is invalid
        * the "test" user belongs to group "group"
        * all other users have no groups

    """
    if "bad" in userid:
        return None
    if userid == "test":
        return ["group"]
    return []


def stub_view_public(request):
    """Stub view that returns userid if logged in, None otherwise."""
    userid = unauthenticated_userid(request)
    return Response(str(userid))


def stub_view_auth(request):
    """Stub view that returns userid if logged in, fails if not."""
    userid = authenticated_userid(request)
    if userid is None:
        raise HTTPForbidden
    return Response(userid)


def stub_view_groups(request):
    """Stub view that returns groups if logged in, fails if not."""
    groups = effective_principals(request)
    return Response(json.dumps([str(g) for g in groups]))

def stub_decode_mac_id(request, id, suffix="-SECRET"):
    """Stub mac-id-decoding function that appends suffix to give the secret."""
    return id, id + suffix


def stub_encode_mac_id(request, id, suffix="-SECRET"):
    """Stub mac-id-encoding function that appends suffix to give the secret."""
    return id, id + suffix

def make_claims(userid, claims=None):
    if claims is None:
        claims = {}
    if 'sub' not in claims:
        claims['sub'] = userid
    now = datetime.datetime.utcnow()
    if 'iat' not in claims:
        claims['iat'] = now
    if 'nbf' not in claims:
        claims['nbf'] = now
    if 'exp' not in claims:
        claims['exp'] = now + datetime.timedelta(seconds=10)
    return claims


class TestJWTAuthenticationPolicy(unittest.TestCase):
    """Testcases for the JWTAuthenticationPolicy class."""

    def setUp(self):
        self.config = Configurator(settings={
            "jwtauth.find_groups": "pyramid_jwtauth.tests.test_jwtauth:stub_find_groups",
            "jwtauth.master_secret": MASTER_SECRET,
        })
        self.config.include("pyramid_jwtauth")
        self.config.add_route("public", "/public")
        self.config.add_view(stub_view_public, route_name="public")
        self.config.add_route("auth", "/auth")
        self.config.add_view(stub_view_auth, route_name="auth")
        self.config.add_route("groups", "/groups")
        self.config.add_view(stub_view_groups, route_name="groups")
        self.app = TestApp(self.config.make_wsgi_app())
        self.policy = self.config.registry.queryUtility(IAuthenticationPolicy)

    def _make_request(self, *args, **kwds):
        return make_request(self.config, *args, **kwds)

    # create an authenticated request that has a JWT with the userid set
    # in our case, this is the 'sub' request.
    #
    # we need to have the following claims as a minimum:
    # 'sub': the userid that we want to authenticate - it can be anything.
    # 'iat': the issued at time stamp
    # 'nbf': the not before time stamp
    # 'exp': the expiry time for the JWT
    def _make_authenticated_request(self, userid, *args, **kwds):
        claims = None
        if 'claims' in kwds:
            claims = kwds['claims']
            del kwds['claims']
        req = self._make_request(*args, **kwds)
        # creds = self._get_credentials(req, userid=userid)
        claims = make_claims(userid=userid, claims=claims)
        # jwt_authenticate_request(req, **creds)
        # note jwt_authenticate_request() returns headers if wanted
        jwt_authenticate_request(req, claims, MASTER_SECRET)
        return req

    def test_the_class_implements_auth_policy_interface(self):
        verifyClass(IAuthenticationPolicy, JWTAuthenticationPolicy)

    def test_from_settings_can_explicitly_set_all_properties(self):
        policy = JWTAuthenticationPolicy.from_settings({
          "jwtauth.find_groups": "pyramid_jwtauth.tests.test_jwtauth:stub_find_groups",
          "jwtauth.master_secret": MASTER_SECRET,
          # "jwtauth.decode_mac_id": "pyramid_macauth.tests:stub_decode_mac_id",
          # "jwtauth.encode_mac_id": "pyramid_macauth.tests:stub_encode_mac_id",
        })
        self.assertEqual(policy.find_groups, stub_find_groups)
        self.assertEqual(policy.master_secret, MASTER_SECRET)
        # self.assertTrue(isinstance(policy.nonce_cache, macauthlib.NonceCache))
        # self.assertEqual(policy.decode_mac_id, stub_decode_mac_id)
        # self.assertEqual(policy.encode_mac_id, stub_encode_mac_id)

    # def test_from_settings_passes_on_args_to_nonce_cache(self):
    #     policy = MACAuthenticationPolicy.from_settings({
    #       "macauth.nonce_cache": "macauthlib:NonceCache",
    #       "macauth.nonce_cache_nonce_ttl": 42,
    #     })
    #     self.assertTrue(isinstance(policy.nonce_cache, macauthlib.NonceCache))
    #     self.assertEqual(policy.nonce_cache.nonce_ttl, 42)
    #     self.assertRaises(TypeError, MACAuthenticationPolicy.from_settings, {
    #       "macauth.nonce_cache": "macauthlib:NonceCache",
    #       "macauth.nonce_cache_invalid_arg": "WHAWHAWHAWHA",
    #     })

    def test_from_settings_errors_out_on_unexpected_keyword_args(self):
        self.assertRaises(ValueError, JWTAuthenticationPolicy.from_settings, {
          "jwtauth.unexpected": "spanish-inquisition",
        })

    # def test_from_settings_errors_out_on_args_to_a_non_callable(self):
    #     self.assertRaises(ValueError, MACAuthenticationPolicy.from_settings, {
    #       "macauth.nonce_cache": "pyramid_macauth.tests:stub_non_callable",
    #       "macauth.nonce_cache_arg": "invalidarg",
    #     })

    # def test_from_settings_errors_out_if_decode_mac_id_is_not_callable(self):
    #     self.assertRaises(ValueError, MACAuthenticationPolicy.from_settings, {
    #       "macauth.decode_mac_id": "pyramid_macauth.tests:stub_non_callable",
    #     })

    # def test_from_settings_errors_out_if_encode_mac_id_is_not_callable(self):
    #     self.assertRaises(ValueError, MACAuthenticationPolicy.from_settings, {
    #       "macauth.encode_mac_id": "pyramid_macauth.tests:stub_non_callable",
    #     })

    def test_from_settings_produces_sensible_defaults(self):
        policy = JWTAuthenticationPolicy.from_settings({})
        # Using __code__ here is a Py2/Py3 compatible way of checking
        # that a bound and unbound method point to the same function object.
        self.assertEqual(policy.find_groups.__code__,
                          JWTAuthenticationPolicy.find_groups.__code__)

    # def test_from_settings_curries_args_to_decode_mac_id(self):
    #     policy = MACAuthenticationPolicy.from_settings({
    #       "macauth.decode_mac_id": "pyramid_macauth.tests:stub_decode_mac_id",
    #       "macauth.decode_mac_id_suffix": "-TEST",
    #     })
    #     self.assertEqual(policy.decode_mac_id(None, "id"), ("id", "id-TEST"))

    # def test_from_settings_curries_args_to_encode_mac_id(self):
    #     policy = MACAuthenticationPolicy.from_settings({
    #       "macauth.encode_mac_id": "pyramid_macauth.tests:stub_encode_mac_id",
    #       "macauth.encode_mac_id_suffix": "-TEST",
    #     })
    #     self.assertEqual(policy.encode_mac_id(None, "id"), ("id", "id-TEST"))

    # not sure if this test is particularly useful as JWTAuthenticationPolicy
    # doesn't work on 'remembering' requests.
    def test_remember_does_nothing(self):
        policy = JWTAuthenticationPolicy()
        req = self._make_authenticated_request("test@moz.com", "/")
        self.assertEqual(policy.remember(req, "test@moz.com"), [])

    def test_forget_gives_a_challenge_header(self):
        policy = JWTAuthenticationPolicy()
        req = self._make_authenticated_request("test@moz.com", "/")
        headers = policy.forget(req)
        self.assertEqual(len(headers), 1)
        self.assertEqual(headers[0][0], "WWW-Authenticate")
        self.assertTrue(headers[0][1] == "JWT")

    def test_forget_gives_a_challenge_header_with_custom_scheme(self):
        policy = JWTAuthenticationPolicy(scheme='Bearer')
        req = self._make_authenticated_request("test@moz.com", "/")
        headers = policy.forget(req)
        self.assertTrue(headers[0][1] == "Bearer")

    def test_unauthenticated_requests_get_a_challenge(self):
        r = self.app.get("/auth", status=401)
        challenge = r.headers["WWW-Authenticate"]
        self.assertTrue(challenge.startswith("JWT"))

    def test_unauthenticated_requests_get_a_challenge_custom_scheme(self):
        self.policy.scheme = 'Bearer'
        r = self.app.get("/auth", status=401)
        challenge = r.headers["WWW-Authenticate"]
        self.assertTrue(challenge.startswith('Bearer'))

    def test_authenticated_request_works(self):
        req = self._make_authenticated_request("test@moz.com", "/auth")
        r = self.app.request(req)
        self.assertEqual(r.body, b"test@moz.com")

    # def test_authentication_fails_when_macid_has_no_userid(self):
    #     req = self._make_request("/auth")
    #     creds = self._get_credentials(req, hello="world")
    #     macauthlib.sign_request(req, **creds)
    #     self.app.request(req, status=401)

    def test_authentication_with_non_jwt_scheme_fails(self):
        req = self._make_request("/auth")
        req.authorization = "OpenID hello=world"
        self.app.request(req, status=401)
        req = self._make_request("/public")
        req.authorization = "OpenID hello=world"
        self.app.request(req, status=200)

    # def test_authentication_without_macid_fails(self):
    #     req = self._make_signed_request("test@moz.com", "/auth")
    #     authz = req.environ["HTTP_AUTHORIZATION"]
    #     authz = authz.replace("id", "idd")
    #     req.environ["HTTP_AUTHORIZATION"] = authz
    #     self.app.request(req, status=401)

    # def test_authentication_without_timestamp_fails(self):
    #     req = self._make_signed_request("test@moz.com", "/auth")
    #     authz = req.environ["HTTP_AUTHORIZATION"]
    #     authz = authz.replace("ts", "typostamp")
    #     req.environ["HTTP_AUTHORIZATION"] = authz
    #     self.app.request(req, status=401)

    # def test_authentication_without_nonce_fails(self):
    #     req = self._make_signed_request("test@moz.com", "/auth")
    #     authz = req.environ["HTTP_AUTHORIZATION"]
    #     authz = authz.replace("nonce", "typonce")
    #     req.environ["HTTP_AUTHORIZATION"] = authz
    #     self.app.request(req, status=401)

    def test_authentication_with_expired_timestamp_fails(self):
        req = self._make_authenticated_request("test@moz.com",
                                               "/auth")
        self.app.request(req, status=200)
        # now do an out of date one.
        ts = datetime.datetime.utcnow() - datetime.timedelta(days=1)
        claims = {'exp': ts}
        req = self._make_authenticated_request("test@moz.com",
                                               "/auth",
                                               claims=claims)
        self.app.request(req, status=401)

    def test_can_get_claims_from_token(self):
        claims = {
            'urn:websandhq.co.uk/auth:jti': 'hello'
        }
        req = self._make_authenticated_request("test@moz.com",
                                               "/auth",
                                               claims=claims)
        policy = JWTAuthenticationPolicy(
            master_secret="V8 JUICE IS 1/8TH GASOLINE")
        encoded_claims = policy.get_claims(req)
        self.assertTrue('urn:websandhq.co.uk/auth:jti' in encoded_claims)
        self.assertEqual(encoded_claims['urn:websandhq.co.uk/auth:jti'],
                          'hello')


    # def test_authentication_with_far_future_timestamp_fails(self):
    #     req = self._make_request("/auth")
    #     creds = self._get_credentials(req, username="test@moz.com")
    #     # Do an initial request so that the server can
    #     # calculate and cache our clock skew.
    #     ts = str(int(time.time()))
    #     req.authorization = ("MAC", {"ts": ts})
    #     macauthlib.sign_request(req, **creds)
    #     self.app.request(req, status=200)
    #     # Now do one with a far future timestamp.
    #     ts = str(int(time.time() + 1000))
    #     req.authorization = ("MAC", {"ts": ts})
    #     macauthlib.sign_request(req, **creds)
    #     self.app.request(req, status=401)

    # def test_authentication_with_reused_nonce_fails(self):
    #     req = self._make_request("/auth")
    #     creds = self._get_credentials(req, username="test@moz.com")
    #     # First request with that nonce should succeed.
    #     req.authorization = ("MAC", {"nonce": "PEPPER"})
    #     macauthlib.sign_request(req, **creds)
    #     r = self.app.request(req)
    #     self.assertEqual(r.body, b"test@moz.com")
    #     # Second request with that nonce should fail.
    #     req = self._make_request("/auth")
    #     req.authorization = ("MAC", {"nonce": "PEPPER"})
    #     macauthlib.sign_request(req, **creds)
    #     self.app.request(req, status=401)

    # def test_authentication_with_busted_macid_fails(self):
    #     req = self._make_signed_request("test@moz.com", "/auth")
    #     id = macauthlib.utils.parse_authz_header(req)["id"]
    #     authz = req.environ["HTTP_AUTHORIZATION"]
    #     authz = authz.replace(id, "XXX" + id)
    #     req.environ["HTTP_AUTHORIZATION"] = authz
    #     self.app.request(req, status=401)

    # def test_authentication_with_busted_signature_fails(self):
    #     req = self._make_request("/auth")
    #     creds = self._get_credentials(req, username="test@moz.com")
    #     macauthlib.sign_request(req, **creds)
    #     signature = macauthlib.utils.parse_authz_header(req)["mac"]
    #     authz = req.environ["HTTP_AUTHORIZATION"]
    #     authz = authz.replace(signature, "XXX" + signature)
    #     req.environ["HTTP_AUTHORIZATION"] = authz
    #     self.app.request(req, status=401)

    def test_groupfinder_can_block_authentication(self):
        req = self._make_authenticated_request("baduser", "/auth")
        r = self.app.request(req, status=403)
        req = self._make_authenticated_request("baduser", "/public")
        r = self.app.request(req, status=200)
        self.assertEqual(r.body, b"baduser")

    def test_groupfinder_groups_are_correctly_reported(self):
        req = self._make_request("/groups")
        r = self.app.request(req)
        self.assertEqual(r.json,
                          [str(Everyone)])
        req = self._make_authenticated_request("gooduser", "/groups")
        r = self.app.request(req)
        self.assertEqual(r.json,
                          ["gooduser", str(Everyone), str(Authenticated)])
        req = self._make_authenticated_request("test", "/groups")
        r = self.app.request(req)
        self.assertEqual(r.json,
                          ["test", str(Everyone), str(Authenticated), "group"])
        req = self._make_authenticated_request("baduser", "/groups")
        r = self.app.request(req)
        self.assertEqual(r.json,
                          [str(Everyone)])

    def test_access_to_public_urls(self):
        # Request with no credentials is allowed access.
        req = self._make_request("/public")
        resp = self.app.request(req)
        self.assertEqual(resp.body, b"None")
        # Request with valid credentials is allowed access.
        req = self._make_authenticated_request("test@moz.com", "/public")
        resp = self.app.request(req)
        self.assertEqual(resp.body, b"test@moz.com")
        # Request with invalid credentials still reports a userid.
        req = self._make_authenticated_request("test@moz.com", "/public")
        token = pyramid_jwtauth.utils.parse_authz_header(req)["token"]
        bits = token.split('.')
        bits[2] = 'XXX' + bits[2]
        broken_token = '.'.join(bits)
        authz = req.environ["HTTP_AUTHORIZATION"]
        authz = authz.replace(token, broken_token)
        req.environ["HTTP_AUTHORIZATION"] = authz
        resp = self.app.request(req)
        self.assertEqual(resp.body, b"test@moz.com")
        # Request with malformed credentials gets a 401
        # req = self._make_signed_request("test@moz.com", "/public")
        # tokenid = macauthlib.utils.parse_authz_header(req)["id"]
        # authz = req.environ["HTTP_AUTHORIZATION"]
        # authz = authz.replace(tokenid, "XXX" + tokenid)
        # req.environ["HTTP_AUTHORIZATION"] = authz
        # resp = self.app.request(req, status=401)

    # def test_check_signature_fails_if_no_params_present(self):
    #     req = self._make_request("/auth")
    #     self.assertRaises(HTTPUnauthorized,
    #                       self.policy._check_signature, req, "XXX")

    def test_default_groupfinder_returns_empty_list(self):
        policy = JWTAuthenticationPolicy()
        req = self._make_request("/auth")
        self.assertEqual(policy.find_groups("test", req), [])

    def test_auth_can_be_checked_several_times_on_same_request(self):
        req = self._make_authenticated_request("test@moz.com", "/public")
        self.assertEqual(authenticated_userid(req), "test@moz.com")
        self.assertEqual(authenticated_userid(req), "test@moz.com")


class TestJWTAuthenticationPolicy_with_RSA(unittest.TestCase):
    """Testcases for the JWTAuthenticationPolicy class."""

    def setUp(self):
        self.config = Configurator(settings={
          "jwtauth.private_key_file": 'pyramid_jwtauth/tests/testkey',
          "jwtauth.public_key_file": 'pyramid_jwtauth/tests/testkey.pub',
          "jwtauth.algorithm": "RS256",
        })
        self.config.include("pyramid_jwtauth")
        self.config.add_route("public", "/public")
        self.config.add_view(stub_view_public, route_name="public")
        self.config.add_route("auth", "/auth")
        self.config.add_view(stub_view_auth, route_name="auth")
        self.config.add_route("groups", "/groups")
        self.config.add_view(stub_view_groups, route_name="groups")
        self.app = TestApp(self.config.make_wsgi_app())
        self.policy = self.config.registry.queryUtility(IAuthenticationPolicy)

    def _make_request(self, *args, **kwds):
        return make_request(self.config, *args, **kwds)

    def test_from_settings_with_RSA_public_private_key(self):
        self.assertEqual(self.policy.algorithm, 'RS256')
        self.assertEqual(self.policy.master_secret, None)
        # from Crypto.PublicKey import RSA
        with open('pyramid_jwtauth/tests/testkey', 'r') as rsa_priv_file:
            private_key = rsa_priv_file.read()
            self.assertEqual(self.policy.private_key, private_key)
        with open('pyramid_jwtauth/tests/testkey.pub', 'r') as rsa_pub_file:
            public_key = rsa_pub_file.read()
            self. assertEqual(self.policy.public_key, public_key)

        self.assertNotEqual(private_key, public_key)
        req = self._make_request("/auth")
        claims = make_claims(userid="test@moz.com")
        jwt_authenticate_request(req, claims, private_key, 'RS256')

        token = pyramid_jwtauth.utils.parse_authz_header(req)["token"]
        userid = self.policy.authenticated_userid(req)
        self.assertEqual(userid, "test@moz.com")

        import jwt
        payload = jwt.decode(token, key=public_key, verify=True)
        self.assertIn('sub', payload)
        self.assertEqual(payload['sub'], "test@moz.com")

        r = self.app.request(req)
        self.assertEqual(r.body, b"test@moz.com")


class TestJWTAuthenticationPolicyJWTOptions(unittest.TestCase):

    def _make_request(self, *args, **kwds):
        return make_request(self.config, *args, **kwds)

    def setup_helper(self, config_settings):
        self.config = Configurator(settings=config_settings)
        self.config.include("pyramid_jwtauth")
        # self.config.add_route("public", "/public")
        # self.config.add_view(stub_view_public, route_name="public")
        # self.config.add_route("auth", "/auth")
        # self.config.add_view(stub_view_auth, route_name="auth")
        # self.config.add_route("groups", "/groups")
        # self.config.add_view(stub_view_groups, route_name="groups")
        self.app = TestApp(self.config.make_wsgi_app())
        self.policy = self.config.registry.queryUtility(IAuthenticationPolicy)


    def test_decode_fails_no_override_on_aud_claim(self):
        self.setup_helper({
            "jwtauth.find_groups": "pyramid_jwtauth.tests.test_jwtauth:stub_find_groups",
            "jwtauth.master_secret": MASTER_SECRET,})
        # make a token with an audience claim
        claims = make_claims('user1', claims={'aud': 'me'})
        req = self._make_request("/auth")
        jwt_authenticate_request(req, claims, MASTER_SECRET)
        token = pyramid_jwtauth.utils.parse_authz_header(req)["token"]
        from jwt.exceptions import InvalidAudienceError
        with self.assertRaises(InvalidAudienceError):
            userid = self.policy.authenticated_userid(req)
        # self.assertEqual(userid, "test@moz.com")

    def test_decode_passes_with_override_on_aud_claim(self):
        self.setup_helper({
            "jwtauth.find_groups": "pyramid_jwtauth.tests.test_jwtauth:stub_find_groups",
            "jwtauth.master_secret": MASTER_SECRET,
            "jwtauth.disable_verify_aud": True,})
        # make a token with an audience claim
        claims = make_claims('user1', claims={'aud': 'me'})
        req = self._make_request("/auth")
        jwt_authenticate_request(req, claims, MASTER_SECRET)
        token = pyramid_jwtauth.utils.parse_authz_header(req)["token"]
        userid = self.policy.authenticated_userid(req)
        self.assertEqual(userid, "user1")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
"""

A Pyramid authentication plugin for JSON Web Tokens:

    http://self-issued.info/docs/draft-ietf-oauth-json-web-token.html

"""

from __future__ import absolute_import

__ver_major__ = 0
__ver_minor__ = 1
__ver_patch__ = 2
__ver_sub__ = ""
__ver_tuple__ = (__ver_major__, __ver_minor__, __ver_patch__, __ver_sub__)
__version__ = "%d.%d.%d%s" % __ver_tuple__

import sys
import functools

from datetime import datetime
from calendar import timegm

from zope.interface import implementer

from pyramid.interfaces import IAuthenticationPolicy
from pyramid.security import Everyone, Authenticated
from pyramid.authorization import ACLAuthorizationPolicy
from pyramid.httpexceptions import HTTPUnauthorized
from pyramid.util import DottedNameResolver

import jwt

from .utils import parse_authz_header, normalize_request_object


@implementer(IAuthenticationPolicy)
class JWTAuthenticationPolicy(object):
    """Pyramid Authentication Policy implementing JWT Access Auth.

    This class provides an IAuthenticationPolicy implementation based on
    signed requests, using the JSON Web Token Authentication standard.

    The plugin can be customized with the following arguments:

        * find_groups:  a callable taking a userid and a Request object, and
                        returning a list of the groups that userid is a
                        member of.

        * master_secret:  a secret known only by the server, used for signing
                          JWT auth tokens in the default implementation.

        * private_key:  An RSA private_key
        * private_key_file: a file holding an RSA encoded (PEM/DER) key file.

        * public_key:  An RSA public_key
        * public_key_file: a file holding an RSA encoded (PEM/DER) key file.

        * algorithm:  The algorithm used to sign the key (defaults to HS256)

        * leeway:  The default leeway (as a datetime.timedelta). Defaults to
                   None

        * userid_in_claim: The claim that the userid can be found in.  Normally
                           this is the 'sub' claim of the JWT, but this can
                           be overridden here.  This is used in
                           authenticated_userid() and related functions.

        * scheme: The scheme name used in the ``Authorization`` header. JWT
          implementations vary in their use of ``JWT`` (our default) or
          ``Bearer``.

    The following configuration options are to DISABLE the verification options
    in the PyJWT decode function.  If the app configures this then it OUGHT to
    ensure that the claim is verified IN the application.

        * decode_options (these are passed to the __init__() = undefined OR {}
          with the following keys (these are the defaults):

            options = {
               'verify_signature': True,
               'verify_exp': True,
               'verify_nbf': True,
               'verify_iat': True,
               'verify_aud': True
            }

          i.e to switch off audience checking, pass 'verify_aud': True in
          decode_options.

          These are passed as the following as part of the ini options/settings

          jwtauth.disable_verify_signature = true (default false)
          jwtauth.disable_verify_exp = true (default false)
          jwtauth.disable_verify_nbf = true (default false)
          jwtauth.disable_verify_iat = true (default false)
          jwtauth.disable_verify_aud = true (default false)

          NOTE: they are reversed between the settings vs the __init__().

    The library takes either a master_secret or private_key/public_key pair.
    In the later case the algorithm must be an RS* version.
    """

    # The default value of master_secret is None, which will cause the library
    # to generate a fresh secret at application startup.
    master_secret = None

    def __init__(self,
                 find_groups=None,
                 master_secret=None,
                 private_key=None,
                 private_key_file=None,
                 public_key=None,
                 public_key_file=None,
                 algorithm='HS256',
                 leeway=None,
                 userid_in_claim=None,
                 scheme='JWT',
                 decode_options=None):
        if find_groups is not None:
            self.find_groups = find_groups
        if master_secret is not None:
            self.master_secret = master_secret
        self.private_key = private_key
        if private_key_file is not None:
            with open(private_key_file, 'r') as rsa_priv_file:
                self.private_key = rsa_priv_file.read()
        self.public_key = public_key
        if public_key_file is not None:
            with open(public_key_file, 'r') as rsa_pub_file:
                self.public_key = rsa_pub_file.read()
        self.algorithm = algorithm
        if leeway is not None:
            self.leeway = leeway
        else:
            self.leeway = 0
        if userid_in_claim is not None:
            self.userid_in_claim = userid_in_claim
        else:
            self.userid_in_claim = 'sub'
        self.scheme = scheme
        self.decode_options = decode_options

    @classmethod
    def from_settings(cls, settings={}, prefix="jwtauth.", **extra):
        """Construct a JWTAuthenticationPolicy from deployment settings.

        This is a helper function for loading a JWTAuthenticationPolicy from
        settings provided in the pyramid application registry.  It extracts
        settings with the given prefix, converts them to the appropriate type
        and passes them into the constructor.
        """
        # Grab out all the settings keys that start with our prefix.
        jwtauth_settings = {}
        for name in settings:
            if not name.startswith(prefix):
                continue
            jwtauth_settings[name[len(prefix):]] = settings[name]
        # Update with any additional keyword arguments.
        jwtauth_settings.update(extra)
        # Pull out the expected keyword arguments.
        kwds = cls._parse_settings(jwtauth_settings)
        # Error out if there are unknown settings.
        for unknown_setting in jwtauth_settings:
            raise ValueError("unknown jwtauth setting: %s" % unknown_setting)
        # And finally we can finally create the object.
        return cls(**kwds)

    @classmethod
    def _parse_settings(cls, settings):
        """Parse settings for an instance of this class.

        This classmethod takes a dict of string settings and parses them into
        a dict of properly-typed keyword arguments, suitable for passing to
        the default constructor of this class.

        Implementations should remove each setting from the dict as it is
        processesed, so that any unsupported settings can be detected by the
        calling code.
        """
        load_function = _load_function_from_settings
        # load_object = _load_object_from_settings
        kwds = {}
        kwds["find_groups"] = load_function("find_groups", settings)
        kwds["master_secret"] = settings.pop("master_secret", None)
        kwds["private_key"] = settings.pop("private_key", None)
        kwds["private_key_file"] = settings.pop("private_key_file", None)
        kwds["public_key"] = settings.pop("public_key", None)
        kwds["public_key_file"] = settings.pop("public_key_file", None)
        kwds["algorithm"] = settings.pop("algorithm", "HS256")
        kwds["leeway"] = settings.pop("leeway", 0)
        kwds["userid_in_claim"] = settings.pop("userid_in_claim", "sub")
        kwds["scheme"] = settings.pop("scheme", "JWT")
        disable_options = {
            'verify_signature': settings.pop("disable_verify_signature", None),
            'verify_exp': settings.pop("disable_verify_exp", None),
            'verify_nbf': settings.pop("disable_verify_nbf", None),
            'verify_iat': settings.pop("disable_verify_iat", None),
            'verify_aud': settings.pop("disable_verify_aud", None),
        }
        kwds["decode_options"] = {
            k: not v for k, v in disable_options.items()}
        return kwds

    def authenticated_userid(self, request):
        """Get the authenticated userid for the given request.

        This method extracts the claimed userid from the request, checks
        the request signature in _get_credentials(), and calls the groupfinder
        callback to check the validity of the claimed identity.
        """
        userid = self._get_credentials(request)
        self._check_signature(request)
        if userid is None:
            return None
        if self.find_groups(userid, request) is None:
            return None
        return userid

    def unauthenticated_userid(self, request):
        """Get the unauthenticated userid for the given request.

        This method extracts the claimed userid from the request without
        checking its authenticity.  This means that the request signature
        is *not* checked when you call this method.  The groupfinder
        callback is also not called.
        """
        userid = self._get_credentials(request)
        return userid

    def effective_principals(self, request):
        """Get the list of effective principals for the given request.

        This method combines the authenticated userid from the request with
        with the list of groups returned by the groupfinder callback, if any.
        """
        principals = [Everyone]
        userid = self._get_credentials(request)
        if userid is None:
            return principals
        self._check_signature(request)
        groups = self.find_groups(userid, request)
        if groups is None:
            return principals
        principals.insert(0, userid)
        principals.append(Authenticated)
        principals.extend(groups)
        return principals

    def remember(self, request, principal, **kw):
        """Get headers to remember to given principal identity.

        This is a no-op for this plugin; the client is supposed to remember
        its MAC credentials and use them for all requests.
        """
        return []

    def forget(self, request):
        """Get headers to forget the identity in the given request.

        This simply issues a new WWW-Authenticate challenge, which should
        cause the client to forget any previously-provisioned credentials.
        """
        return [("WWW-Authenticate", self.scheme)]

    def challenge(self, request, content="Unauthorized"):
        """Challenge the user for credentials.

        This method returns a 401 response using the WWW-Authenticate field
        as constructed by forget().  You might like to use it as pyramid's
        "forbidden view" when using this auth policy.
        """
        return HTTPUnauthorized(content, headers=self.forget(request))

    def find_groups(self, userid, request):
        """Find the list of groups for the given userid.

        This method provides a default implementation of the "groupfinder
        callback" used by many pyramid authn policies to look up additional
        user data.  It can be overridden by passing a callable into the
        JWTAuthenticationPolicy constructor.

        The default implementation returns an empty list.
        """
        return []

    def decode_jwt(self, request, jwtauth_token,
                   leeway=None, verify=True, options=None):
        """Decode a JWTAuth token into its claims.

        This method deocdes the given JWT to provide the claims.  The JWT can
        fail if the token has expired (with appropriate leeway) or if the
        token won't validate due to the secret (key) being wrong.

        If the JWT doesn't verify then a number of Exceptions can be raised:
            DecodeError() - if the algorithm in the token isn't supported.
            DecodeError() - if the secret doesn't match (key, etc.)
            ExpiredSignature() - if the 'exp' claim has expired.

        If private_key/public key is set then the public_key will be used to
        decode the key.

        Note that the 'options' value is normally None, as this function is
        usually called via the (un)authenticated_userid() which is called by
        the framework.  Thus the decode 'options' are set as part of
        configuring the module through Pyramid settings.

        :param request: the Pyramid Request object
        :param jwtauth_token: the string (bString - Py3) - of the full token
                              to decode
        :param leeway: Integer - the number of seconds of leeway to pass to
                       jwt.decode()
        :param verify: Boolean - True to verify - passed to jwt.decode()
        :param options: set of options for what to verify.
        """
        if leeway is None:
            leeway = self.leeway
        if self.public_key is not None:
            key = self.public_key
        else:
            key = self.master_secret
        _options = self.decode_options or {}
        if options:
            _options.update(options)
        if len(_options.keys()) == 0:
            _options = None
        claims = jwt.decode(jwtauth_token,
                            key=key,
                            leeway=leeway,
                            verify=verify,
                            options=_options)
        return claims

    def encode_jwt(self, request, claims, key=None, algorithm=None):
        """Encode a set of claims into a JWT token.

        This is just a proxy for jwt.encode() but uses the default
        master_secret that may have been set in configuring the library.

        If the private_key is set then self.private_key is used for the encode
        (assuming key = None!)  algorithm also has to be an RS* algorithm and
        if not set, then self.algorithm is used.
        """
        if key is None:
            if self.private_key is not None:
                key = self.private_key
            else:
                key = self.master_secret
        if algorithm is None:
            algorithm = self.algorithm
        # fix for older version of PyJWT which doesn't covert all of the time
        # claims.  This won't be needed in the future.
        encode_claims = maybe_encode_time_claims(claims)

        jwtauth_token = jwt.encode(encode_claims, key=key, algorithm=algorithm)
        return jwtauth_token

    def _get_params(self, request):
        """Get the JWTAuth parameters from the given request.

        This method parses the Authorization header to get the JSON Web
        Token. If they seem sensible, we cache them in the request
        to avoid reparsing and return them as a dict.

        If the request contains no JWT Auth credentials, None is returned.
        """
        try:
            return request.environ["jwtauth.params"]
        except KeyError:
            params = parse_authz_header(request, None)
            if params is not None:
                if params.get("scheme").upper() !=self.scheme:
                    params = None
            request.environ["jwtauth.params"] = params
            return params

    def _get_credentials(self, request):
        """Extract the JWTAuth claims from the request.

        This method extracts and returns the claimed userid from the JWTAuth
        data in the request, along with the corresonding request signing
        key.  It does *not* check the signature on the request.

        If there are no JWTAuth credentials in the request then None
        is returned.
        """
        userid = request.environ.get("jwtauth.userid", False)
        if userid:
            return userid

        params = self._get_params(request)
        if params is None:
            return None
        if 'token' not in params:
            return None
        # Now try to pull out the claims from the JWT - note it is unusable if
        # we get a decode error, but might be okay if we get a signature error
        # Thus we may have to call decode TWICE, once with verify=True to see
        # if we just get a jwt.ExpiredSignature or jwt.DecodeError and if so,
        # the second time with verify=False to try to get the claims (i.e. to
        # ignore the jwt.ExpiredSignature)  we store whether the signature is
        # okay in jwtauth.signature_is_valid environ on the request
        def _get_claims():
            try:
                claims = self.decode_jwt(request, params['token'], verify=True)
                return claims, True
            except (jwt.DecodeError, jwt.ExpiredSignature):
                # try again with no verify
                try:
                    claims = self.decode_jwt(
                        request, params['token'], verify=False)
                    return claims, False
                except jwt.DecodeError:
                    # can't do anything with this.
                    return None, False

        claims, verify_okay = _get_claims()
        if claims is None:
            return None
        # so we don't have to check it again.
        request.environ["jwtauth.claims"] = claims
        request.environ["jwtauth.signature_is_valid"] = verify_okay
        # Now extract the userid
        if self.userid_in_claim in claims:
            request.environ["jwtauth.userid"] = claims[self.userid_in_claim]
            return claims[self.userid_in_claim]
        return None

    def get_claims(self, request):
        """Get the claims from the request - if they exist.

        Fetch the claims out of the token on the request, if it exists and is
        decodable.  Returns None if there are none or it couldn't be docoded.
        """
        userid = self._get_credentials(request)
        if userid is None:
            return None
        return request.environ.get("jwtauth.claims", None)

    def _check_signature(self, request):
        """See if the signature was valid

        It was already checked in _get_credentials() - this function just
        sees if it was valid.

        If the JWTAuth token id is invalid then HTTPUnauthorized
        will be raised.
        """
        # See if we've already checked the signature on this request.
        # This is important because pyramid doesn't cache the results
        # of authenticating the request, but we mark the nonce as stale
        # after the first check.
        if request.environ.get("jwtauth.signature_is_valid", False):
            return True
        # Grab the (hopefully cached) params from the request.
        params = self._get_params(request)
        if params is None:
            msg = "missing JWT token"
            raise self.challenge(request, msg)
        # We know the JWT auth token's signature isn't valid:
        msg = "invalid JWT signature"
        raise self.challenge(request, msg)


def maybe_encode_time_claims(claims):
    encode_claims = claims.copy()
    # convert datetime to a intDate value in known time-format claims
    for time_claim in ['exp', 'iat', 'nbf']:
        if isinstance(encode_claims.get(time_claim), datetime):
            encode_claims[time_claim] = (
                timegm(encode_claims[time_claim].utctimetuple()))
    return encode_claims


@normalize_request_object
def authenticate_request(request, claims, key, algorithm='HS256', scheme='JWT'):
    """Authenticate a webob style request with the appropriate JWT token

    This creates the auth token using the claims and the key to ensure that
    will be accepted by this library.  Obviously, normally, a client would be
    making the request - so this is just useful as a 'canonical' way of
    creating a Authorization header
    """
    claims = maybe_encode_time_claims(claims)
    jwtauth_token = jwt.encode(claims, key=key, algorithm=algorithm)
    if sys.version_info >= (3, 0, 0):
        jwtauth_token = jwtauth_token.decode(encoding='UTF-8')
    params = dict()
    params['token'] = jwtauth_token
    # Serialize the parameters back into the authz header, and return it.
    # WebOb has logic to do this that's not perfect, but good enough for us.
    request.authorization = (scheme, params)
    return request.headers['Authorization']


def _load_function_from_settings(name, settings):
    """Load a plugin argument as a function created from the given settings.

    This function is a helper to load and possibly curry a callable argument
    to the plugin.  It grabs the value from the dotted python name found in
    settings[name] and checks that it is a callable.  It then looks for args
    of the form settings[name_*] and curries them into the function as extra
    keyword argument before returning.
    """
    # See if we actually have the named object.
    dotted_name = settings.pop(name, None)
    if dotted_name is None:
        return None
    func = DottedNameResolver(None).resolve(dotted_name)
    # Check that it's a callable.
    if not callable(func):
        raise ValueError("Argument %r must be callable" % (name,))
    # Curry in any keyword arguments.
    func_kwds = {}
    prefix = name + "_"
    for key in list(settings.keys()):
        if key.startswith(prefix):
            func_kwds[key[len(prefix):]] = settings.pop(key)
    # Return the original function if not currying anything.
    # This is both more efficent and better for unit testing.
    if func_kwds:
        func = functools.partial(func, **func_kwds)
    return func


def _load_object_from_settings(name, settings):
    """Load a plugin argument as an object created from the given settings.

    This function is a helper to load and possibly instanciate an argument
    to the plugin.  It grabs the value from the dotted python name found in
    settings[name].  If this is a callable, it looks for arguments of the
    form settings[name_*] and calls it with them to instanciate an object.
    """
    # See if we actually have the named object.
    dotted_name = settings.pop(name, None)
    if dotted_name is None:
        return None
    obj = DottedNameResolver(None).resolve(dotted_name)
    # Extract any arguments for the callable.
    obj_kwds = {}
    prefix = name + "_"
    for key in list(settings.keys()):
        if key.startswith(prefix):
            obj_kwds[key[len(prefix):]] = settings.pop(key)
    # Call it if callable.
    if callable(obj):
        obj = obj(**obj_kwds)
    elif obj_kwds:
        raise ValueError("arguments provided for non-callable %r" % (name,))
    return obj


def includeme(config):
    """Install JWTAuthenticationPolicy into the provided configurator.

    This function provides an easy way to install JWT Access Authentication
    into your pyramid application.  Loads a JWTAuthenticationPolicy from the
    deployment settings and installs it into the configurator.
    """
    # Hook up a default AuthorizationPolicy.
    # ACLAuthorizationPolicy is usually what you want.
    # If the app configures one explicitly then this will get overridden.
    # In auto-commit mode this needs to be set before adding an authn policy.
    authz_policy = ACLAuthorizationPolicy()
    config.set_authorization_policy(authz_policy)

    # Build a JWTAuthenticationPolicy from the deployment settings.
    settings = config.get_settings()
    authn_policy = JWTAuthenticationPolicy.from_settings(settings)
    config.set_authentication_policy(authn_policy)

    # Set the forbidden view to use the challenge() method on the policy.
    # The following causes a problem with cornice (fighting - open to options
    # about them playing properly together.)
    # config.add_forbidden_view(authn_policy.challenge)




import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'README.rst')) as f:
    README = f.read()

with open(os.path.join(here, 'CHANGES.txt')) as f:
    CHANGES = f.read()

# TODO: fix the imports.
requires = ['pyramid >= 1.3', 'webtest', 'cryptography', 'PyJWT']

setup(name='pyramid_jwtauth',
      version='0.1.3',
      description='pyramid_jwtauth',
      long_description=README + '\n\n' + CHANGES,
      license='MPLv2.0',
      classifiers=[
        "Development Status :: 4 - Beta",
        "Framework :: Pyramid",
        "Intended Audience :: Developers",
        "Topic :: Internet :: WWW/HTTP",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        ],
      author='Alex Kavanagh (Websand)',
      author_email='alex@ajkavanagh.co.uk',
      url='https://github.com/ajkavanagh/pyramid_jwtauth',
      keywords='authentication token JWT JSON',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      install_requires=requires,
      tests_require=requires,
      test_suite="pyramid_jwtauth")
